
using System;
using System.Collections.Generic;
using Dashboard.ImmediateDraw;
using OpenTK.Mathematics;

namespace Dashboard.Controls
{

    /// <summary>
    /// Bases for all UI elements.
    /// </summary>
    public abstract class UIBase : IDbAttribute
    {
        private Vector2 size;

        public UIBase? Parent { get; protected set; }
        public Rectangle Bounds
        {
            get => new Rectangle(Position + Size, Position);
            set
            {
                Size = value.Size;
                Position = value.Min;
            }
        }

        public virtual Vector2 Position { get; set; }

        public Vector2 Size
        {
            get => size;
            set
            {
                Vector2 oldSize = size;
                size = value;

                OnResized(this, new ResizedEventArgs(size, oldSize));
            }
        }

        public Rectangle AbsoluteBounds
        {
            get
            {
                if (Parent == null)
                {
                    return Bounds;
                }
                else
                {
                    return new Rectangle(Bounds.Max + Parent.Position, Bounds.Min + Parent.Position);
                }
            }
        }

        public Vector2 MaximumSize { get; set; } = new Vector2(-1, -1);
        public Vector2 MinimumSize { get; set; } = new Vector2(-1, -1);

        public bool IsMaximumSizeSet => MaximumSize != new Vector2(-1, -1);
        public bool IsMinimumSizeSet => MinimumSize != new Vector2(-1, -1);

        public Dictionary<string, object> Attributes { get; } = new Dictionary<string, object>();

        public virtual void NotifyEvent(object? sender, EventArgs args)
        {
        }

        protected virtual void PaintBegin(DrawList cmd)
        {
            cmd.PushViewport();
            cmd.StoreViewport(AbsoluteBounds);
            cmd.PushZ();
        }

        protected virtual void PaintEnd(DrawList cmd)
        {
            cmd.PopViewport();
        }

        public void Paint(DrawList cmd)
        {
            PaintBegin(cmd);
            PaintEnd(cmd);
        }

        public event EventHandler<ResizedEventArgs>? Resized;
        public event EventHandler<AdoptedEventArgs>? Adopted;

        public virtual void OnResized(object sender, ResizedEventArgs ea)
        {
            Resized?.Invoke(sender, ea);
        }

        protected virtual void OnAdopted(UIBase? parent)
        {
            Adopted?.Invoke(this, new AdoptedEventArgs(parent));
        }

        public bool IsDisposed { get; private set; } = false;

        protected virtual void Dispose(bool disposing)
        {
            foreach (object userdata in Attributes.Values)
            {
                if (userdata is IDisposable disposable)
                    disposable.Dispose();
            }
        }

        protected void DisposeInvoker(bool disposing)
        {
            if (IsDisposed)
                return;

            IsDisposed = true;

            Dispose(disposing);
        }

        public void Dispose() => DisposeInvoker(true);

        protected static void Adopt(UIBase child, UIBase? parent)
        {
            child.Parent = parent;
            child.OnAdopted(parent);
        }
    }

    public class ResizedEventArgs : EventArgs
    {
        public Vector2 NewSize { get; }
        public Vector2 OldSize { get; }

        public ResizedEventArgs(Vector2 newSize, Vector2 oldSize)
        {
            NewSize = newSize;
            OldSize = oldSize;
        }
    }

    public class AdoptedEventArgs : EventArgs
    {
        public UIBase? Parent { get; }

        public AdoptedEventArgs(UIBase? parent = null)
        {
            Parent = parent;
        }
    }
}